# JUBE Benchmarking Environment
# Copyright (C) 2008-2019
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Fileset related classes"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

import os
import shutil
import xml.etree.ElementTree as ET
import jube2.util.util
import jube2.conf
import jube2.step
import jube2.log
import glob

LOGGER = jube2.log.get_logger(__name__)


class Fileset(list):

    """Container for file copy, link and prepare operations"""

    def __init__(self, name):
        list.__init__(self)
        self._name = name

    @property
    def name(self):
        """Return fileset name"""
        return self._name

    def etree_repr(self):
        """Return etree object representation"""
        fileset_etree = ET.Element("fileset")
        fileset_etree.attrib["name"] = self._name
        for file_handle in self:
            fileset_etree.append(file_handle.etree_repr())
        return fileset_etree

    def create(self, work_dir, parameter_dict, alt_work_dir=None,
               environment=None, file_path_ref=""):
        """Copy/load/prepare all files in fileset"""
        for file_handle in self:
            if type(file_handle) is Prepare:
                file_handle.execute(
                    parameter_dict=parameter_dict,
                    work_dir=alt_work_dir if alt_work_dir
                    is not None else work_dir,
                    environment=environment)
            else:
                file_handle.create(
                    work_dir=work_dir, parameter_dict=parameter_dict,
                    alt_work_dir=alt_work_dir, file_path_ref=file_path_ref,
                    environment=environment)


class File(object):

    """Generic file access"""

    def __init__(self, path, name=None, is_internal_ref=False, active="true",
                 source_dir="", target_dir=""):
        self._path = path
        self._source_dir = source_dir
        self._target_dir = target_dir
        self._name = name
        self._file_path_ref = ""
        self._active = active
        self._is_internal_ref = is_internal_ref

    def create(self, work_dir, parameter_dict, alt_work_dir=None,
               file_path_ref="", environment=None):
        """Create file access"""
        # Check active status
        active = jube2.util.util.eval_bool(jube2.util.util.substitution(
            self._active, parameter_dict))
        if not active:
            return
        pathname = jube2.util.util.substitution(self._path, parameter_dict)
        pathname = os.path.expanduser(pathname)
        source_dir = jube2.util.util.substitution(self._source_dir,
                                                  parameter_dict)
        source_dir = os.path.expanduser(source_dir)
        target_dir = jube2.util.util.substitution(self._target_dir,
                                                  parameter_dict)
        target_dir = os.path.expanduser(target_dir)
        if environment is not None:
            pathname = jube2.util.util.substitution(pathname, environment)
            source_dir = jube2.util.util.substitution(source_dir, environment)
            target_dir = jube2.util.util.substitution(target_dir, environment)
        else:
            pathname = os.path.expandvars(pathname)
            source_dir = os.path.expandvars(source_dir)
            target_dir = os.path.expandvars(target_dir)

        # Add source prefix directory if needed
        pathname = os.path.join(source_dir, pathname)

        if self._is_internal_ref:
            pathname = os.path.join(work_dir, pathname)
        else:
            pathname = os.path.join(self._file_path_ref, pathname)
            pathname = os.path.join(file_path_ref, pathname)
            pathname = os.path.normpath(pathname)
        if self._name is None:
            name = os.path.basename(pathname)
        else:
            name = jube2.util.util.substitution(self._name, parameter_dict)
            name = os.path.expanduser(name)
            if environment is not None:
                name = jube2.util.util.substitution(name, environment)
            else:
                name = os.path.expandvars(name)

        if alt_work_dir is not None:
            work_dir = alt_work_dir
        # Shell expansion
        pathes = glob.glob(pathname)
        if (len(pathes) == 0) and (not jube2.conf.DEBUG_MODE):
            raise RuntimeError("no files found using \"{0}\""
                               .format(pathname))
        for path in pathes:
            # When using shell extensions, alternative filenames are not
            # allowed for multiple matches.
            if (len(pathes) > 1) or ((pathname != path) and
                                     (name == os.path.basename(pathname))):
                name = os.path.basename(path)

            # Add target prefix directory if needed
            name = os.path.join(target_dir, name)

            new_file_path = os.path.join(work_dir, name)

            # Create target_dir if needed
            if (len(os.path.dirname(new_file_path)) > 0 and
                    not os.path.exists(os.path.dirname(new_file_path)) and
                    not jube2.conf.DEBUG_MODE):
                os.makedirs(os.path.dirname(new_file_path))

            self.create_action(path, name, new_file_path)

    def create_action(self, path, name, new_file_path):
        """File access type specific creation"""
        raise NotImplementedError()

    def etree_repr(self):
        """Return etree object representation"""
        raise NotImplementedError()

    @property
    def path(self):
        """Return filepath"""
        return self._path

    @property
    def file_path_ref(self):
        """Get file path reference"""
        return self._file_path_ref

    @file_path_ref.setter
    def file_path_ref(self, file_path_ref):
        """Set file path reference"""
        self._file_path_ref = file_path_ref

    @property
    def is_internal_ref(self):
        """Return path is internal ref"""
        return self._is_internal_ref

    def __repr__(self):
        return self._path


class Link(File):

    """A link to a given path. Which can be used inside steps."""

    def create_action(self, path, name, new_file_path):
        """Create link to file in work_dir"""
        # Manipulate target_path if a new relative name path was selected
        if os.path.isabs(path):
            target_path = path
        else:
            target_path = os.path.relpath(path, os.path.dirname(new_file_path))
        LOGGER.debug("  link \"{0}\" <- \"{1}\"".format(target_path, name))
        if not jube2.conf.DEBUG_MODE and not os.path.exists(new_file_path):
            os.symlink(target_path, new_file_path)

    def etree_repr(self):
        """Return etree object representation"""
        link_etree = ET.Element("link")
        link_etree.text = self._path
        if self._name is not None:
            link_etree.attrib["name"] = self._name
        if self._active != "true":
            link_etree.attrib["active"] = self._active
        if self._source_dir != "":
            link_etree.attrib["source_dir"] = self._source_dir
        if self._target_dir != "":
            link_etree.attrib["target_dir"] = self._target_dir
        if self._is_internal_ref:
            link_etree.attrib["rel_path_ref"] = "internal"
        if self._file_path_ref != "":
            link_etree.attrib["file_path_ref"] = self._file_path_ref
        return link_etree


class Copy(File):

    """A file or directory given by path. Which can be copied to the work_dir
    inside steps.
    """

    def create_action(self, path, name, new_file_path):
        """Copy file/directory to work_dir"""
        LOGGER.debug("  copy \"{0}\" -> \"{1}\"".format(path, name))
        if not jube2.conf.DEBUG_MODE and not os.path.exists(new_file_path):
            if os.path.isdir(path):
                shutil.copytree(path, new_file_path, symlinks=True)
            else:
                shutil.copy2(path, new_file_path)

    def etree_repr(self):
        """Return etree object representation"""
        copy_etree = ET.Element("copy")
        copy_etree.text = self._path
        if self._name is not None:
            copy_etree.attrib["name"] = self._name
        if self._active != "true":
            copy_etree.attrib["active"] = self._active
        if self._source_dir != "":
            copy_etree.attrib["source_dir"] = self._source_dir
        if self._target_dir != "":
            copy_etree.attrib["target_dir"] = self._target_dir
        if self._is_internal_ref:
            copy_etree.attrib["rel_path_ref"] = "internal"
        if self._file_path_ref != "":
            copy_etree.attrib["file_path_ref"] = self._file_path_ref
        return copy_etree


class Prepare(jube2.step.Operation):

    """Prepare the workpackage work directory"""

    def __init__(self, cmd, stdout_filename=None, stderr_filename=None,
                 work_dir=None, active="true"):
        jube2.step.Operation.__init__(self,
                                      do=cmd,
                                      stdout_filename=stdout_filename,
                                      stderr_filename=stderr_filename,
                                      active=active,
                                      work_dir=work_dir)

    def execute(self, parameter_dict, work_dir, only_check_pending=False,
                environment=None):
        """Execute the prepare command"""
        jube2.step.Operation.execute(
            self, parameter_dict=parameter_dict, work_dir=work_dir,
            only_check_pending=only_check_pending, environment=environment)

    def etree_repr(self):
        """Return etree object representation"""
        do_etree = ET.Element("prepare")
        do_etree.text = self._do
        if self._stdout_filename is not None:
            do_etree.attrib["stdout"] = self._stdout_filename
        if self._stderr_filename is not None:
            do_etree.attrib["stderr"] = self._stderr_filename
        if self._active != "true":
            do_etree.attrib["active"] = self._active
        if self._work_dir is not None:
            do_etree.attrib["work_dir"] = self._work_dir
        return do_etree



# JUBE Benchmarking Environment
# Copyright (C) 2008-2019
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Resulttype definition"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

import jube2.util.util
import xml.etree.ElementTree as ET
import re
import jube2.log

LOGGER = jube2.log.get_logger(__name__)


class Result(object):

    """A generic result type"""

    class ResultData(object):

        """A gerneric result data type"""

        def __init__(self, name):
            self._name = name

        @property
        def name(self):
            """Return the result name"""
            return self._name

        def create_result(self, show=True, filename=None, **kwargs):
            """Create result output"""
            raise NotImplementedError("")

        def add_result_data(self, result_data):
            """Add additional result data"""
            raise NotImplementedError("")

        def __eq__(self, other):
            return self.name == other.name

    def __init__(self, name, res_filter=None):
        self._use = set()
        self._name = name
        self._res_filter = res_filter
        self._result_dir = None
        self._benchmark = None

    @property
    def name(self):
        """Return the result name"""
        return self._name

    @property
    def benchmark(self):
        """Return the benchmark"""
        return self._benchmark

    @property
    def result_dir(self):
        """Return the result_dir"""
        return self._result_dir

    @result_dir.setter
    def result_dir(self, result_dir):
        """Set the result_dir"""
        self._result_dir = result_dir

    @benchmark.setter
    def benchmark(self, benchmark):
        """Set the benchmark"""
        self._benchmark = benchmark

    def add_uses(self, use_names):
        """Add an addtional analyser name"""
        for use_name in use_names:
            if use_name in self._use:
                raise ValueError(("Element \"{0}\" can only be used once")
                                 .format(use_name))
            self._use.add(use_name)

    def create_result_data(self):
        """Create result representation"""
        raise NotImplementedError("")

    def _analyse_data(self):
        """Load analyse data out of given analysers"""
        for analyser_name in self._use:
            analyser = self._benchmark.analyser[analyser_name]
            analyse = analyser.analyse_result
            # Ignore empty analyse results
            if analyse is None:
                LOGGER.warning(("No data found for analyser \"{0}\" "
                                "in benchmark run {1}. "
                                "Run analyse step first please.")
                               .format(analyser_name, self._benchmark.id))
                continue

            # Create workpackage chains
            wp_chains = list()
            all_wps = set()
            for ids in [analyse[stepname].keys() for stepname in analyse]:
                all_wps.update(set(map(int, ids)))
            # Create copy of all wps to reduce this list while wps are sorted
            # into the chains
            all_wps_tmp = all_wps.copy()

            while (len(all_wps_tmp) > 0):
                next_id = all_wps_tmp.pop()
                # Create new chain
                wp_chains.append(list())
                # Add all parents to the chain
                for wp in self._benchmark.workpackage_by_id(next_id).\
                        parent_history:
                    if wp.id in all_wps:
                        wp_chains[-1].append(wp.id)
                        all_wps_tmp.discard(wp.id)
                # Add wp itself to the cahin
                wp_chains[-1].append(next_id)
                steps_in_list = set()
                # Add children to the chain, each step can only be added once)
                for wp in self._benchmark.workpackage_by_id(next_id).\
                        children_future:
                    if wp.step.name not in steps_in_list:
                        if wp.id in all_wps:
                            steps_in_list.add(wp.step.name)
                            wp_chains[-1].append(wp.id)
                            all_wps_tmp.discard(wp.id)

            # Add all parent WPs which might be missing in the chain, by
            # using the last WP as a reference. This WPs does not provide any
            # analyse data but, their parameter will be available
            for i, chain in enumerate(wp_chains):
                for wp in self._benchmark.workpackage_by_id(chain[-1]).\
                        parent_history:
                    if wp.id not in chain:
                        wp_chains[i] = [wp.id] + chain

            # Create output datasets by combining analyse and parameter data
            for chain in wp_chains:
                analyse_dict = dict()
                for wp_id in chain:
                    workpackage = self._benchmark.workpackage_by_id(wp_id)
                    # add analyse data
                    if (wp_id in all_wps):
                        analyse_dict.update(
                            analyse[workpackage.step.name][wp_id])
                    # add parameter
                    parameter_dict = dict()
                    for par in workpackage.parameterset:
                        value = \
                            jube2.util.util.convert_type(par.parameter_type,
                                                         par.value, stop=False)
                        # add suffix to the parameter name
                        if (par.name + "_" + workpackage.step.name
                                not in parameter_dict):
                            parameter_dict[par.name + "_" +
                                           workpackage.step.name] = value
                        # parmater without suffix si used for the last WP in
                        # the chain
                        if wp_id == chain[-1]:
                            parameter_dict[par.name] = value
                    analyse_dict.update(parameter_dict)
                    # Add jube additional information
                    analyse_dict.update({
                        "jube_res_analyser": analyser_name,
                    })

                # If res_filter is set, only show matching result lines
                if self._res_filter is not None:
                    res_filter = jube2.util.util.substitution(
                        self._res_filter, analyse_dict)
                    if not jube2.util.util.eval_bool(res_filter):
                        continue

                yield analyse_dict

    def _load_units(self, pattern_names):
        """Load units"""
        units = dict()

        alt_pattern_names = list(pattern_names)
        for i, pattern_name in enumerate(alt_pattern_names):
            for option in ["last", "min", "max", "avg", "sum", "std"]:
                matcher = re.match("^(.+)_{0}$".format(option), pattern_name)
                if matcher:
                    alt_pattern_names[i] = matcher.group(1)

        for analyser_name in self._use:
            if analyser_name not in self._benchmark.analyser:
                raise RuntimeError(
                    "<analyser name=\"{0}\"> not found".format(analyser_name))
            patternset_names = \
                self._benchmark.analyser[analyser_name].use.copy()
            for analyse_files in \
                    self._benchmark.analyser[analyser_name].analyser.values():
                for analyse_file in analyse_files:
                    for use in analyse_file.use:
                        patternset_names.add(use)
            for patternset_name in patternset_names:
                patternset = self._benchmark.patternsets[patternset_name]
                for i, pattern_name in enumerate(pattern_names):
                    alt_pattern_name = alt_pattern_names[i]
                    if (pattern_name in patternset) or \
                            (alt_pattern_name in patternset):
                        pattern = patternset[pattern_name]
                        if pattern is None:
                            pattern = patternset[alt_pattern_name]
                        if (pattern.unit is not None) and (pattern.unit != ""):
                            units[pattern_name] = pattern.unit
        return units

    def etree_repr(self):
        """Return etree object representation"""
        result_etree = ET.Element("result")
        if self._result_dir is not None:
            result_etree.attrib["result_dir"] = self._result_dir
        for use in self._use:
            use_etree = ET.SubElement(result_etree, "use")
            use_etree.text = use
        return result_etree



#!/usr/bin/env python
from __future__ import print_function
import sys
import os
import re
import datetime 

maxSpaceDim = 15 # max No. of steps in one JUQUEEN job
defaultmaxWallclock=datetime.datetime.strptime("06:00:00","%H:%M:%S") # max Wallclock time for 1 step
smallmaxWallclock=datetime.datetime.strptime("00:30:00","%H:%M:%S") # max Wallclock time for 1 step (node <=64)
class Job:
    globalJobSpec = []
    def __init__(self,jubeStep,pwd):
        submitFile=os.path.join("..",jubeStep,"work","submit.job")
        try:
            f=open(submitFile,"r")
        except IOError:
            print('ERROR: unable to open submit script "{0}"'.format(submitFile), 
                  file=sys.stderr)  
            raise
        self.name=jubeStep
        self.pwd=pwd
        self._parseJobscript(f)
        f.close()

    def _parseJobscript(self,fptr):
        globalJobSpec=[]
        specificJobSpec=[]
        jobCommands=[]
        stdoutfile="stdout"
        errorfile="stderr"
        walltime="00:01:00"  
        for line in fptr:
            if re.findall(r"^#\s?@\s*(shell|job_name|notification|class|notify_user)",line):
                globalJobSpec.append(line)
            elif re.findall(r"^#\s?@\s*output\s*",line):
                stdoutfile=re.findall(r"^#\s?@\s*output\s*=\s*(\S+)\s*$",line)[0]
            elif re.findall(r"^#\s?@\s*error\s*",line):
                errorfile=re.findall(r"^#\s?@\s*error\s*=\s*(\S+)\s*$",line)[0]
            elif re.findall(r"^#\s?@\s*wall_clock_limit\s*",line):
                walltime=re.findall(r"^#\s?@\s*wall_clock_limit\s*=\s*(\S+)\s*$",line)[0]
            elif re.findall(r"^#+\s?@\s*",line):
                specificJobSpec.append(line)
            elif re.findall(r"^[ #\n\t\r\f]*$",line):
                continue
            else:
                jobCommands.append(line)
        self.__class__.globalJobSpec=globalJobSpec
        self.specificJobSpec=specificJobSpec
        self.jobCommands=jobCommands
        self.stdout=stdoutfile
        self.stderr=errorfile
        self.walltime=datetime.datetime.strptime(walltime,"%H:%M:%S")
        return 

def writeJobHead():
    return [ "#@ output = $(job_name).$(step_name).$(jobid)\n",
             "#@ error = $(output)\n",
             "#@ environment = COPY_ALL\n"
         ]

def initJobStep(step):
    return [ "#=== Step {0:d} directives ===\n".format(step),
             "#@ step_name = step_{0:d}\n".format(step),
             "#@ dependency = (step_{0:d} >= 0)\n".format(step-1) if step > 0 else ""
         ]
 
def writeJobCommands(job,cwd):
    stdout = os.path.join(job.pwd,job.stdout)
    stderr = os.path.join(job.pwd,job.stderr)
    output  = [ '(cd {0}\n'.format(job.pwd) ]
    output += job.jobCommands
    output += [ 'cd {0}\n'.format(cwd) ]
    output += [ ') > {0} 2> {1}\n'.format(stdout,stderr) ]
    return output
    

################################
#   MAIN
################################

filename = sys.argv[1]

f=open(filename,"r")
SystemParameterSpace={}
for line in f:
    lineArray=line.split(':')
    jubeStep = lineArray[0]
    jubeStepPWD = lineArray[3]
    SystemParameter=(lineArray[1],lineArray[2]) # nodes,topology
    if SystemParameter in SystemParameterSpace:
        SystemParameterSpace[SystemParameter].append(Job(jubeStep,jubeStepPWD))
    else:
        SystemParameterSpace[SystemParameter]=[Job(jubeStep,jubeStepPWD)]

# Abort if it is not possible to create the chained jobs
if len(SystemParameterSpace) > maxSpaceDim:
    print('ERROR: system specific parameter space to big', file=sys.stderr)  
    print('       allowed dimension: {0:3d}'.format(maxSpaceDim), file=sys.stderr)  
    print('       used dimension:    {0:3d}'.format(len(SystemParameterSpace)), file=sys.stderr)  
    for (index,parameter) in enumerate(SystemParameterSpace.keys()):
        print('       {0:d}. {1}'.format(index+1,str(parameter)), file=sys.stderr)
    exit(1)
f.close()
        
zeroTime=datetime.datetime.strptime("00:00:00","%H:%M:%S") 
commandlines=["case $LOADL_STEP_NAME in\n"]
cwd = os.environ['PWD']
step = 0
fptr = open("submit.job","w")
fptr.writelines(Job.globalJobSpec)
fptr.writelines(writeJobHead())
for SystemParameter in SystemParameterSpace.keys():
    substep=0
    maxWallclock = defaultmaxWallclock if int(SystemParameter[0])>64 else smallmaxWallclock
    stepWalltime=datetime.timedelta(0) # init with  0
    commandlines += ['step_{0:d}) echo "Working on $LOADL_STEP_NAME"\n'.format(step)]
    for job in SystemParameterSpace[SystemParameter]:
        if job.walltime  > maxWallclock:
            print('ERROR: wall_clock_limit "{0}" exceeded by one job specification - ABORTED'.format(job.walltime), 
                  file=sys.stderr)  
            exit(2)
        elif job.walltime + stepWalltime > maxWallclock:
            fptr.writelines(initJobStep(step))
            fptr.writelines(job.specificJobSpec)
            fptr.write("#@ wall_clock_limit    = {0}\n\n".format((zeroTime + stepWalltime).strftime("%H:%M:%S"))) 
            stepWalltime = job.walltime-zeroTime
            substep = 0
            step += 1
            commandlines += [';;\n\n','step_{0:d}) echo "Working on $LOADL_STEP_NAME"\n'.format(step)]
        else:
            substep += 1
            stepWalltime += job.walltime-zeroTime
        commandlines += writeJobCommands(job,cwd)
    commandlines += [';;\n\n'] 
    fptr.writelines(initJobStep(step))
    fptr.writelines(job.specificJobSpec)
    fptr.write("#@ wall_clock_limit    = {0}\n\n".format((zeroTime + stepWalltime).strftime("%H:%M:%S")))            
    step += 1
commandlines += ["esac\n"]
fptr.writelines(commandlines)
fptr.close()

if step-1 > maxSpaceDim:
    print('ERROR: cumulated wall_clock_limit to big', file=sys.stderr)  
    print('       => exceeded max step limitation: {0:3d}'.format(maxSpaceDim), file=sys.stderr)  
    exit(3)

exit(0)



# JUBE Benchmarking Environment
# Copyright (C) 2008-2019
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Configuration"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

# general
JUBE_VERSION = "2.3.0"
ALLOWED_SCRIPTTYPES = set(["python", "perl", "shell"])
ALLOWED_ADVANCED_MODETYPES = set(["tag", "env"])
ALLOWED_MODETYPES = set(["text"]).union(ALLOWED_SCRIPTTYPES).union(
    ALLOWED_ADVANCED_MODETYPES)
DEBUG_MODE = False
VERBOSE_LEVEL = 0
UPDATE_VERSION_URL = "http://apps.fz-juelich.de/jsc/jube/jube2/version"
UPDATE_URL = "http://apps.fz-juelich.de/jsc/jube/jube2/download.php"
STANDARD_SHELL = "/bin/sh"
EXIT_ON_ERROR = False

# input/output
DEFAULT_SEPARATOR = ","
ZERO_FILL_DEFAULT = 6
DEFAULT_WIDTH = 70
MAX_TABLE_CELL_WIDTH = 40
HIDE_ANIMATIONS = False
VERBOSE_STDOUT_READ_CHUNK_SIZE = 50
VERBOSE_STDOUT_POLL_SLEEP = 0.05
SYSLOG_FMT_STRING = "jube[%(process)s]: %(message)s"
PREPROCESS_MAX_ITERATION = 10

# filenames
WORKPACKAGE_DONE_FILENAME = "done"
WORKPACKAGE_ERROR_FILENAME = "error"
CONFIGURATION_FILENAME = "configuration.xml"
WORKPACKAGES_FILENAME = "workpackages.xml"
ANALYSE_FILENAME = "analyse.xml"
RESULT_DIRNAME = "result"
ENVIRONMENT_INFO = "jube_environment_information.dat"
TIMESTAMPS_INFO = "timestamps"

# logging
DEFAULT_LOGFILE_NAME = "jube-parse.log"
LOGFILE_DEBUG_NAME = "jube-debug.log"
LOGFILE_DEBUG_MODE = "w"
LOGFILE_RUN_NAME = "run.log"
LOGFILE_CONTINUE_NAME = "continue.log"
LOGFILE_ANALYSE_NAME = "analyse.log"
LOGFILE_PARSE_NAME = "parse.log"
LOGFILE_RESULT_NAME = "result.log"
LOG_CONSOLE_FORMAT = "%(message)s"
LOG_FILE_FORMAT = "[%(asctime)s]:%(levelname)s: %(message)s"
DEFAULT_LOGGING_MODE = "default"

# other
ERROR_MSG_LINES = 5
MAX_RECURSIVE_SUB = 5



# JUBE Benchmarking Environment
# Copyright (C) 2008-2019
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Logging Support"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

import logging
import sys
import glob
import os.path
import jube2.conf


class JubeLogger(logging.getLoggerClass(), object):

    """Overwrite logging to handle multi line messages."""

    def _log(self, level, msg, *args, **kwargs):
        """Log multi line messages each as a separate entry."""
        if hasattr(msg, "splitlines"):
            lines = msg.splitlines()
        else:
            lines = str(msg).splitlines()
        for line in lines:
            super(JubeLogger, self)._log(level, line, *args, **kwargs)


logging.setLoggerClass(JubeLogger)

LOGGING_MODE = jube2.conf.DEFAULT_LOGGING_MODE
LOGFILE_NAME = jube2.conf.DEFAULT_LOGFILE_NAME
CONSOLE_VERBOSE = False


def get_logger(name=None):
    """Return logger given by name"""
    return logging.getLogger(name)


def setup_logging(mode=None, filename=None, verbose=None):
    """Setup the logging configuration.

    Available modes are
      default   log to console and file
      console   only console output

    filename can be given optionally.

    verbose: enable verbose console output

    The setup includes setting the handlers and formatters. Calling
    this function multiple times causes old handlers to be removed
    before new ones are added.

    """
    global LOGGING_MODE, LOGFILE_NAME, CONSOLE_VERBOSE

    # Use debug file name and debug file mode when in debug mode
    if jube2.conf.DEBUG_MODE:
        filename = jube2.conf.LOGFILE_DEBUG_NAME
        mode = "default"
        filemode = jube2.conf.LOGFILE_DEBUG_MODE
    else:
        filemode = "a"

    if mode is None:
        mode = LOGGING_MODE
    else:
        LOGGING_MODE = mode
    if filename is None:
        filename = LOGFILE_NAME
    else:
        LOGFILE_NAME = filename
    if verbose is None:
        verbose = CONSOLE_VERBOSE
    else:
        CONSOLE_VERBOSE = verbose

    # this is needed to make the other handlers accept on low priority
    # events
    _logger = get_logger("jube2")
    _logger.setLevel(logging.DEBUG)

    # list is needed since we remove from the list we just iterate
    # over
    for handler in list(_logger.handlers):
        handler.close()
        _logger.removeHandler(handler)

    # create, configure and add console handler
    console_formatter = logging.Formatter(jube2.conf.LOG_CONSOLE_FORMAT)
    console_handler = logging.StreamHandler(sys.stdout)
    if verbose:
        console_handler.setLevel(logging.DEBUG)
    else:
        console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(console_formatter)
    _logger.addHandler(console_handler)

    if mode == "default":
        try:
            # create, configure and add file handler
            file_formatter = logging.Formatter(jube2.conf.LOG_FILE_FORMAT)
            file_handler = logging.FileHandler(filename, filemode)
            file_handler.setLevel(logging.DEBUG)
            file_handler.setFormatter(file_formatter)
            _logger.addHandler(file_handler)
        except IOError:
            pass


def search_for_logs(path=None):
    """Search for files matching in path with .log extension"""
    if path is None:
        path = "."
    matches = glob.glob(os.path.join(path, "*.log"))
    return matches


def log_print(text):
    """Output text"""
    print(text)


def matching_logs(commands, available_logs):
    """Find intersection between requested logs and available logs.

    Returns tuple (matching, not_matching), containing the
    intersection and its complement.

    Only compares basenames.

    """
    requested_logs = set("{0}.log".format(command) for command in commands)
    matching = list()
    for log in available_logs:
        if os.path.basename(log) in requested_logs:
            matching.append(log)
    not_matching = requested_logs.difference(set([os.path.basename(log)
                                                  for log in matching]))
    return matching, not_matching


def safe_output_logfile(filename):
    """Try to print logfile. If try fails, fail gracefully."""
    try:
        with open(filename) as logfile:
            log_print(logfile.read())
    except IOError:
        log_print("No log found in current directory")


def change_logfile_name(filename):
    """Change log file name if not in debug mode."""
    if jube2.conf.DEBUG_MODE:
        return
    setup_logging(filename=filename, mode="default")


def only_console_log():
    """Change to console log if not in debug mode."""
    if jube2.conf.DEBUG_MODE:
        return
    setup_logging(mode="console")


def reset_logging():
    """Reset logging to default."""
    global LOGGING_MODE, LOGFILE_NAME

    LOGGING_MODE = jube2.conf.DEFAULT_LOGGING_MODE
    LOGFILE_NAME = jube2.conf.DEFAULT_LOGFILE_NAME



# JUBE Benchmarking Environment
# Copyright (C) 2008-2019
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""jube2.result_types package"""



# JUBE Benchmarking Environment
# Copyright (C) 2008-2019
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""The Analyser class handles the analyse process"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

import xml.etree.ElementTree as ET
import jube2.log
import os
import re
import glob
import math
import jube2.pattern
import jube2.util.util
import jube2.util.output

LOGGER = jube2.log.get_logger(__name__)


class Analyser(object):

    """The Analyser handles the analyse process and store all important data
    to run a new analyse."""

    class AnalyseFile(object):

        """A file which should be analysed"""

        def __init__(self, path):
            self._path = path
            self._use = set()

        def add_uses(self, use_names):
            """Add an addtional patternset name"""
            for use_name in use_names:
                if use_name in self._use:
                    raise ValueError(("Element \"{0}\" can only be used once")
                                     .format(use_name))
                self._use.add(use_name)

        def __eq__(self, other):
            result = len(self._use.symmetric_difference(other.use)) == 0
            return result and (self._path == other.path)

        def __repr__(self):
            return "AnalyseFile({0})".format(self._path)

        @property
        def use(self):
            """Return uses"""
            return self._use

        @property
        def path(self):
            """Get file path"""
            return self._path

        def etree_repr(self):
            """Return etree object representation"""
            file_etree = ET.Element("file")
            file_etree.text = self._path
            if len(self._use) > 0:
                file_etree.attrib["use"] = \
                    jube2.conf.DEFAULT_SEPARATOR.join(self._use)
            return file_etree

    def __init__(self, name, reduce_iteration=True):
        self._name = name
        self._use = set()
        self._analyse = dict()
        self._benchmark = None
        self._analyse_result = None
        self._reduce_iteration = reduce_iteration

    @property
    def benchmark(self):
        """Get benchmark information"""
        return self._benchmark

    @benchmark.setter
    def benchmark(self, benchmark):
        """Set benchmark information"""
        self._benchmark = benchmark

    @property
    def use(self):
        """Return uses"""
        return self._use

    @property
    def analyser(self):
        """Return analyse dict"""
        return self._analyse

    @property
    def analyse_result(self):
        """Return analyse result"""
        return self._analyse_result

    @analyse_result.setter
    def analyse_result(self, analyse_result):
        """Set analyse result"""
        self._analyse_result = analyse_result

    def add_analyse(self, step_name, analyse_file):
        """Add an addtional analyse file"""
        if step_name not in self._analyse:
            self._analyse[step_name] = list()
        if (analyse_file not in self._analyse[step_name]) and \
                (analyse_file is not None):
            self._analyse[step_name].append(analyse_file)

    def add_uses(self, use_names):
        """Add an addtional patternset name"""
        for use_name in use_names:
            if use_name in self._use:
                raise ValueError(("Element \"{0}\" can only be used once")
                                 .format(use_name))
            self._use.add(use_name)

    @property
    def name(self):
        """Get analyser name"""
        return self._name

    def etree_repr(self):
        """Return etree object representation"""
        analyser_etree = ET.Element("analyser")
        analyser_etree.attrib["name"] = self._name
        analyser_etree.attrib["reduce"] = str(self._reduce_iteration)
        for use in self._use:
            use_etree = ET.SubElement(analyser_etree, "use")
            use_etree.text = use
        for step_name in self._analyse:
            analyse_etree = ET.SubElement(analyser_etree, "analyse")
            analyse_etree.attrib["step"] = step_name
            for fileobj in self._analyse[step_name]:
                analyse_etree.append(fileobj.etree_repr())
        return analyser_etree

    def _combine_and_check_patternsets(self, patternset, uses):
        """Combine patternsets given by uses and check compatibility"""
        for use in uses:
            if use not in self._benchmark.patternsets:
                raise RuntimeError(("<patternset name=\"{0}\"> used but not " +
                                    "found").format(use))
            if not patternset.is_compatible(self._benchmark.patternsets[use]):
                incompatible_names = patternset.get_incompatible_pattern(
                    self._benchmark.patternsets[use])
                raise RuntimeError(("Cannot use patternset \"{0}\" " +
                                    "in analyser \"{1}\", because there are " +
                                    "incompatible pattern name combinations: "
                                    "{2}")
                                   .format(use, self._name,
                                           ",".join(incompatible_names)))
            patternset.add_patternset(self._benchmark.patternsets[use])

    def analyse(self):
        """Run the analyser"""
        LOGGER.debug("Run analyser \"{0}\"".format(self._name))
        if self._benchmark is None:
            raise RuntimeError("No benchmark found using analyser {0}"
                               .format(self._name))

        result = dict()

        # Combine all patternsets
        patternset = jube2.pattern.Patternset()
        self._combine_and_check_patternsets(patternset, self._use)

        # Print debug info
        debugstr = "  available pattern:\n"
        debugstr += \
            jube2.util.output.text_table(
                [("pattern", "value")] +
                sorted([(par.name, par.value) for par in
                        patternset.pattern_storage]),
                use_header_line=True, indent=9,
                align_right=False)
        debugstr += "\n  available derived pattern:\n"
        debugstr += \
            jube2.util.output.text_table(
                [("pattern", "value")] +
                sorted([(par.name, par.value) for par in
                        patternset.derived_pattern_storage]),
                use_header_line=True, indent=9,
                align_right=False)
        LOGGER.debug(debugstr)

        for stepname in self._analyse:
            result[stepname] = dict()
            LOGGER.debug("  analyse step \"{0}\"".format(stepname))
            if stepname not in self._benchmark.steps:
                raise RuntimeError(("Could not find <step name=\"{0}\"> "
                                    "when using analyser \"{1}\"").format(
                                        stepname, self._name))
            step = self._benchmark.steps[stepname]
            workpackages = set(self._benchmark.workpackages[stepname])
            while len(workpackages) > 0:
                root_workpackage = workpackages.pop()
                match_dict = dict()
                # Global patternset to store all existing pattern (e.g. from
                # individual file uses), necessary to evaluate default pattern
                # and derived pattern
                global_patternset = patternset.copy()
                result[stepname][root_workpackage.id] = dict()
                # Should multiple iterations be reduced to a single result line
                if self._reduce_iteration:
                    siblings = set(root_workpackage.iteration_siblings)
                else:
                    siblings = set([root_workpackage])
                while len(siblings) > 0:
                    workpackage = siblings.pop()
                    if workpackage in workpackages:
                        workpackages.remove(workpackage)

                    # Ignore workpackages not started yet
                    if not workpackage.started:
                        continue

                    parameter = \
                        dict([[par.name, par.value] for par in
                              workpackage.parameterset.
                              constant_parameter_dict.values()])

                    for file_obj in self._analyse[stepname]:
                        if step.alt_work_dir is not None:
                            file_path = step.alt_work_dir
                            file_path = jube2.util.util.substitution(
                                file_path, parameter)
                            file_path = \
                                os.path.expandvars(
                                    os.path.expanduser(file_path))
                            file_path = os.path.join(
                                self._benchmark.file_path_ref, file_path)
                        else:
                            file_path = workpackage.work_dir

                        filename = \
                            jube2.util.util.substitution(file_obj.path,
                                                         parameter)
                        filename = \
                            os.path.expandvars(os.path.expanduser(filename))

                        file_path = os.path.join(file_path, filename)
                        for path in glob.glob(file_path):
                            # scan files
                            LOGGER.debug(("    scan file {0}").format(path))

                            new_result_dict, match_dict = \
                                self._analyse_file(path, patternset,
                                                   global_patternset,
                                                   workpackage.parameterset,
                                                   match_dict,
                                                   file_obj.use)
                            result[stepname][root_workpackage.id].update(
                                new_result_dict)

                # Set default pattern values if available and necessary
                new_result_dict = result[stepname][root_workpackage.id]
                for pattern in global_patternset.pattern_storage:
                    if (pattern.default_value is not None) and \
                            (pattern.name not in new_result_dict):
                        default = pattern.default_value
                        # Convert default value
                        if pattern.content_type == "int":
                            if default == "nan":
                                default = float("nan")
                            else:
                                default = int(float(default))
                        elif pattern.content_type == "float":
                            default = float(default)
                        new_result_dict[pattern.name] = default
                        new_result_dict[pattern.name + "_cnt"] = 0
                        new_result_dict[pattern.name + "_last"] = default
                        if pattern.content_type in ["int", "float"]:
                            new_result_dict.update(
                                {pattern.name + "_sum": default,
                                 pattern.name + "_min": default,
                                 pattern.name + "_max": default,
                                 pattern.name + "_avg": default,
                                 pattern.name + "_sum2": default ** 2,
                                 pattern.name + "_std": 0})

                # Evaluate derived pattern
                new_result_dict = self._eval_derived_pattern(
                    global_patternset, root_workpackage.parameterset,
                    result[stepname][root_workpackage.id])
                result[stepname][root_workpackage.id].update(
                    new_result_dict)

        self._analyse_result = result

    def _eval_derived_pattern(self, patternset, parameterset, result_dict):
        """Evaluate all derived pattern in patternset using parameterset
        and result_dict"""
        resultset = jube2.parameter.Parameterset()
        for name in result_dict:
            resultset.add_parameter(
                jube2.parameter.Parameter.create_parameter(
                    name, value=str(result_dict[name])))

        # Get jube patternset
        jube_pattern = jube2.pattern.get_jube_pattern()
        # calculate derived pattern
        patternset.derived_pattern_substitution(
            [parameterset, resultset, jube_pattern.pattern_storage])

        new_result_dict = dict()
        # Convert content type
        for par in patternset.derived_pattern_storage:
            if par.mode not in jube2.conf.ALLOWED_SCRIPTTYPES:
                new_result_dict[par.name] = \
                    jube2.util.util.convert_type(par.content_type,
                                                 par.value, stop=False)
        return new_result_dict

    def _analyse_file(self, file_path, patternset, global_patternset,
                      parameterset, match_dict=None, additional_uses=None):
        """Scan given files with given pattern and produce a result
        parameterset"""
        if additional_uses is None:
            additional_uses = set()
        if match_dict is None:
            match_dict = dict()

        if not os.path.isfile(file_path):
            return dict(), match_dict

        local_patternset = patternset.copy()

        # Add file specific uses
        self._combine_and_check_patternsets(local_patternset, additional_uses)
        self._combine_and_check_patternsets(global_patternset, additional_uses)

        # Unique pattern/parameter check
        if (not parameterset.is_compatible(
                local_patternset.pattern_storage)) or \
           (not parameterset.is_compatible(
                local_patternset.derived_pattern_storage)):

            incompatible_names = parameterset.get_incompatible_parameter(
                local_patternset.pattern_storage)
            incompatible_names.update(parameterset.get_incompatible_parameter(
                local_patternset.derived_pattern_storage))
            raise RuntimeError(("A pattern and a parameter (\"{0}\") "
                                "using the same name in "
                                "analyser \"{1}\"").format(
                                    ",".join(incompatible_names), self._name))

        # Get jube patternset
        jube_pattern = jube2.pattern.get_jube_pattern()

        # Do pattern substitution
        local_patternset.pattern_substitution(
            [parameterset, jube_pattern.pattern_storage])

        patternlist = [p for p in local_patternset.pattern_storage]

        file_handle = open(file_path, "r")
        # Read file content
        data = file_handle.read()
        for pattern in patternlist:
            if pattern.name not in match_dict:
                match_dict[pattern.name] = dict()
            try:
                mode = re.MULTILINE
                if pattern.dotall:
                    mode += re.DOTALL
                regex = re.compile(pattern.value, mode)
            except re.error as ree:
                raise RuntimeError(("Error inside pattern \"{0}\" : " +
                                    "\"{1}\" : {2}")
                                   .format(pattern.name, pattern.value, ree))
            # Run regular expression
            matches = re.findall(regex, data)
            # If there are different groups reduce result shape
            if regex.groups > 1:
                match_list = list()
                for match in matches:
                    match_list = match_list + list(match)
            else:
                match_list = matches
            # Remove empty matches
            match_list = [match for match in match_list if match != ""]

            # Convert to pattern type
            new_match_list = list()
            for match in match_list:
                try:
                    if pattern.content_type == "int":
                        if match == "nan":
                            new_match_list.append(float("nan"))
                        else:
                            new_match_list.append(int(float(match)))
                    elif pattern.content_type == "float":
                        new_match_list.append(float(match))
                    else:
                        new_match_list.append(match)
                except ValueError:
                    LOGGER.warning(("\"{0}\" cannot be represented " +
                                    "as a \"{1}\"")
                                   .format(match, pattern.content_type))
            match_list = new_match_list

            if len(match_list) > 0:
                # First match is default
                if "first" not in match_dict[pattern.name]:
                    match_dict[pattern.name]["first"] = match_list[0]

                for match in match_list:
                    if pattern.content_type in ["int", "float"]:
                        if "min" in match_dict[pattern.name]:
                            match_dict[pattern.name]["min"] = \
                                min(match_dict[pattern.name]["min"], match)
                        else:
                            match_dict[pattern.name]["min"] = match
                        if "max" in match_dict[pattern.name]:
                            match_dict[pattern.name]["max"] = \
                                max(match_dict[pattern.name]["max"], match)
                        else:
                            match_dict[pattern.name]["max"] = match
                        if "sum" in match_dict[pattern.name]:
                            match_dict[pattern.name]["sum"] += match
                        else:
                            match_dict[pattern.name]["sum"] = match
                        if "sum2" in match_dict[pattern.name]:
                            match_dict[pattern.name]["sum2"] += match ** 2
                        else:
                            match_dict[pattern.name]["sum2"] = match ** 2

                    if "cnt" in match_dict[pattern.name]:
                        match_dict[pattern.name]["cnt"] += 1
                    else:
                        match_dict[pattern.name]["cnt"] = 1

                if pattern.content_type in ["int", "float"]:
                    if match_dict[pattern.name]["cnt"] > 0:
                        match_dict[pattern.name]["avg"] = \
                            (match_dict[pattern.name]["sum"] /
                             match_dict[pattern.name]["cnt"])

                    if match_dict[pattern.name]["cnt"] > 1:
                        match_dict[pattern.name]["std"] = math.sqrt(
                            (abs(match_dict[pattern.name]["sum2"] -
                                 (match_dict[pattern.name]["sum"] ** 2 /
                                  match_dict[pattern.name]["cnt"])) /
                             (match_dict[pattern.name]["cnt"] - 1)))
                    else:
                        match_dict[pattern.name]["std"] = 0

                match_dict[pattern.name]["last"] = match_list[-1]

        info_str = "      file \"{0}\" scanned pattern found:\n".format(
            os.path.basename(file_path))
        info_str += jube2.util.output.text_table(
            [(_name, ", ".join(["{0}:{1}".format(key, con)
                                for key, con in value.items()]))
             for _name, value in match_dict.items()],
            indent=9, align_right=True, auto_linebreak=True)
        LOGGER.debug(info_str)
        file_handle.close()

        # Create result dict
        result_dict = dict()
        for pattern_name in match_dict:
            for option in match_dict[pattern_name]:
                if option == "first":
                    name = pattern_name
                else:
                    name = "{0}_{1}".format(pattern_name, option)
                result_dict[name] = match_dict[pattern_name][option]

        return result_dict, match_dict

    def analyse_etree_repr(self):
        """Create an etree representation of a analyse dict:
        stepname -> workpackage_id -> filename -> patternname -> value
        """
        etree = list()
        for stepname in self._analyse_result:
            step_etree = ET.Element("step")
            step_etree.attrib["name"] = stepname
            for workpackage_id in self._analyse_result[stepname]:
                workpackage_etree = ET.SubElement(step_etree, "workpackage")
                workpackage_etree.attrib["id"] = str(workpackage_id)
                for pattern in self._analyse_result[stepname][workpackage_id]:
                    if type(self._analyse_result[stepname][workpackage_id]
                            [pattern]) is int:
                        content_type = "int"
                    elif type(self._analyse_result[stepname][
                              workpackage_id][pattern]) is float:
                        content_type = "float"
                    else:
                        content_type = "string"
                    pattern_etree = ET.SubElement(workpackage_etree, "pattern")
                    pattern_etree.attrib["name"] = pattern
                    pattern_etree.attrib["type"] = content_type
                    pattern_etree.text = \
                        str(self._analyse_result[stepname][workpackage_id]
                            [pattern])
            etree.append(step_etree)
        return etree



# JUBE Benchmarking Environment
# Copyright (C) 2008-2019
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Utility functions, constants and classes"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

try:
    import queue
except ImportError:
    import Queue as queue
import re
import string
import operator
import os.path
import subprocess
import jube2.log
import time
import jube2.conf
import grp
import pwd


LOGGER = jube2.log.get_logger(__name__)


class WorkStat(object):

    """Workpackage queuing handler"""

    def __init__(self):
        self._work_list = queue.Queue()
        self._cnt_work = dict()
        self._wait_lists = dict()

    def put(self, workpackage):
        """Add some workpackage to queue"""

        # Substitute max_wps if needed
        max_wps = int(substitution(workpackage.step.max_wps,
                                   workpackage.parameter_dict))

        if (max_wps == 0) or \
           (workpackage.started) or \
           (workpackage.step.name not in self._cnt_work) or \
           (self._cnt_work[workpackage.step.name] < max_wps):
            self._work_list.put(workpackage)
            if workpackage.step.name not in self._cnt_work:
                self._cnt_work[workpackage.step.name] = 1
            else:
                self._cnt_work[workpackage.step.name] += 1
        else:
            if workpackage.step.name not in self._wait_lists:
                self._wait_lists[workpackage.step.name] = queue.Queue()
            self._wait_lists[workpackage.step.name].put(workpackage)

    def update_queues(self, last_workpackage):
        """Check if a workpackage can move from waiting to work queue"""
        if last_workpackage.done:
            self._cnt_work[last_workpackage.step.name] -= 1
            if (last_workpackage.step.name in self._wait_lists) and \
               (not self._wait_lists[last_workpackage.step.name].empty()):
                workpackage = \
                    self._wait_lists[last_workpackage.step.name].get_nowait()
                # Check if workpackage was started from another position
                if not workpackage.started:
                    self.put(workpackage)
                else:
                    self.update_queues(last_workpackage)

    def get(self):
        """Get some workpackage from work queue"""
        return self._work_list.get_nowait()

    def empty(self):
        """Check if work queue is empty"""
        return self._work_list.empty()


def get_current_id(base_dir):
    """Return the highest id found in directory 'base_dir'."""
    try:
        filelist = sorted(os.listdir(base_dir))
    except OSError as error:
        LOGGER.warning(error)
        filelist = list()

    maxi = -1
    for item in filelist:
        try:
            maxi = max(int(re.findall("^([0-9]+)$", item)[0]), maxi)
        except IndexError:
            pass
    return maxi


def id_dir(base_dir, id_number):
    """Return path for 'id_number' in 'base_dir'."""
    return os.path.join(
        base_dir,
        "{id_number:0{zfill}d}".format(zfill=jube2.conf.ZERO_FILL_DEFAULT,
                                       id_number=id_number))


def substitution(text, substitution_dict):
    """Substitute templates given by parameter_dict inside of text"""
    changed = True
    count = 0
    # All values must be string values (handle Python 2 separatly)
    try:
        str_substitution_dict = \
            dict([(k, str(v).decode("utf-8", errors="ignore")) for k, v in
                  substitution_dict.items()])
    except AttributeError:
        str_substitution_dict = dict([(k, str(v)) for k, v in
                                      substitution_dict.items()])
    # Preserve non evaluated parameter before starting substitution
    local_substitution_dict = dict([(k, re.sub(r"\$", "$$", v)
                                     if "$" in v else v) for k, v in
                                    str_substitution_dict.items()])

    # Run multiple times to allow recursive parameter substitution
    while changed and count < jube2.conf.MAX_RECURSIVE_SUB:
        count += 1
        orig_text = text
        # Save double $$
        text = re.sub(r"(\$\$)(?=(\$\$|[^$]))", "$$$$", text) \
            if "$" in text else text
        tmp = string.Template(text)
        new_text = tmp.safe_substitute(local_substitution_dict)
        changed = new_text != orig_text
        text = new_text
    # Final substitution to remove $$
    tmp = string.Template(text)
    return tmp.safe_substitute(str_substitution_dict)


def convert_type(value_type, value, stop=True):
    """Convert value to given type"""
    result_value = None
    try:
        if value_type == "int":
            if value == "nan":
                result_value = float("nan")
            else:
                result_value = int(float(value))
        elif value_type == "float":
            result_value = float(value)
        else:
            result_value = value
    except ValueError:
        if stop:
            raise ValueError(("\"{0}\" cannot be represented as a \"{1}\"")
                             .format(value, value_type))
        else:
            result_value = value
    return result_value


def script_evaluation(cmd, script_type):
    """cmd will be evaluated with given script language"""
    if script_type == "python":
        return str(eval(cmd))
    elif script_type in ["perl", "shell"]:
        if script_type == "perl":
            cmd = "perl -e \"print " + cmd + "\""
        sub = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, shell=True)
        stdout, stderr = sub.communicate()
        stdout = stdout.decode(errors="ignore")
        # Check command execution error code
        errorcode = sub.wait()
        if errorcode != 0:
            raise RuntimeError(stderr)
        else:
            if len(stderr.strip()) > 0:
                try:
                    LOGGER.debug((" The command \"{0}\" was executed with a "
                                  "successful error code,\n  but the "
                                  "following error message was produced "
                                  "during its execution: {1}")
                                 .format(cmd, stderr))
                except UnicodeDecodeError:
                    pass
            return stdout


def eval_bool(cmd):
    """Evaluate a bool expression"""
    if cmd.lower() == "true":
        return True
    elif cmd.lower() == "false":
        return False
    else:
        try:
            return bool(eval(cmd))
        except SyntaxError as se:
            raise ValueError(
                ("\"{0}\" could not be evaluated and handled as boolean "
                 "value. Check if all parameter were correctly replaced and "
                 "the syntax of the expression is well formed ({1}).").format(
                     cmd, str(se)))


def get_tree_element(node, tag_path=None, attribute_dict=None):
    """Can be used instead of node.find(.//tag_path[@attrib=value])"""
    result = get_tree_elements(node, tag_path, attribute_dict)
    if len(result) > 0:
        return result[0]
    else:
        return None


def get_tree_elements(node, tag_path=None, attribute_dict=None):
    """Can be used instead of node.findall(.//tag_path[@attrib=value])"""
    if attribute_dict is None:
        attribute_dict = dict()

    result = list()

    if tag_path is not None:
        node_list = node.findall(tag_path)
    else:
        node_list = [node]

    for found_node in node_list:
        for attribute, value in attribute_dict.items():
            if found_node.get(attribute) != value:
                break
        else:
            result.append(found_node)

    for subtree in node:
        result += get_tree_elements(subtree, tag_path, attribute_dict)

    return result


def now_str():
    """Return current time string"""
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())


def update_timestamps(path, *args):
    """Set all timestamps for given arg_names to now"""
    timestamps = dict()
    timestamps.update(read_timestamps(path))
    file_ptr = open(path, "w")
    for arg in args:
        timestamps[arg] = now_str()
    for timestamp in timestamps:
        file_ptr.write("{0}: {1}\n".format(timestamp, timestamps[timestamp]))
    file_ptr.close()


def read_timestamps(path):
    """Return timestamps dictionary"""
    timestamps = dict()
    if os.path.isfile(path):
        file_ptr = open(path, "r")
        for line in file_ptr:
            matcher = re.match("(.*?): (.*)", line.strip())
            if matcher:
                timestamps[matcher.group(1)] = matcher.group(2)
        file_ptr.close()
    return timestamps


def resolve_depend(depend_dict):
    """Generate a serialization of dependent steps.

    Return a list with a possible order of execution.
    """
    def find_next(dependencies, finished):
        """Returns the next possible items to be processed and remainder.

        dependencies  Dictionary containing the dependencies
        finished      Set which is already processed
        """
        possible = set()
        remain = dict()

        for key, val in dependencies.items():
            if val.issubset(finished):
                possible.add(key)
            else:
                remain[key] = val

        possible.difference_update(finished)
        # no advance
        if dependencies and not possible:
            unresolved_steps = set(dependencies) - finished
            unresolved_dependencies = set()
            for step in unresolved_steps:
                unresolved_dependencies.update(depend_dict[step] -
                                               finished)
            infostr = ("unresolved steps: {0}".
                       format(",".join(unresolved_steps)) + "\n" +
                       "unresolved dependencies: {0}".
                       format(",".join(unresolved_dependencies)))
            LOGGER.warning(infostr)

        return (possible, remain)

    finished = set()
    work_list = list()

    work, remain = find_next(depend_dict, finished)
    while work:
        work_list += list(work)
        finished.update(work)
        work, remain = find_next(remain, finished)

    return work_list


def check_and_get_group_id():
    """Read environment var JUBE_GROUP_NAME and return group id"""
    group_name = ""
    if "JUBE_GROUP_NAME" in os.environ:
        group_name = os.environ["JUBE_GROUP_NAME"].strip()

    if group_name != "":
        try:
            group_id = grp.getgrnam(group_name).gr_gid
        except KeyError:
            raise ValueError(("Failed to get group ID, group \"{0}\" " +
                              "does not exist").format(group_name))
        user = pwd.getpwuid(os.getuid()).pw_name
        grp_members = grp.getgrgid(group_id).gr_mem
        if user in grp_members:
            return group_id
        else:
            raise ValueError(("User \"{0}\" is not in " +
                              "group \"{1}\"").format(user, group_name))
    else:
        return None


def consistency_check(benchmark):
    """Do some consistency checks"""

    # check if step uses exists
    for step in benchmark.steps.values():
        for uses in step.use:
            for use in uses:
                if (use not in benchmark.parametersets) and \
                   (use not in benchmark.filesets) and \
                   (use not in benchmark.substitutesets) and \
                   ("$" not in use):
                    raise ValueError(("<use>{0}</use> not found in "
                                      "available sets").format(use))
    # Dependency check
    depend_dict = \
        dict([(step.name, step.depend) for step in benchmark.steps.values()])
    order = resolve_depend(depend_dict)
    for step_name in benchmark.steps:
        if step_name not in order:
            raise ValueError("Cannot resolve dependencies.")


class CompType(object):
    """Allow comparison of different datatypes"""

    def __init__(self, value):
        self.__value = value

    def __repr__(self):
        return str(self.__value)

    @property
    def value(self):
        return self.__value

    def _special_comp(self, other, comp_func):
        """Allow comparision of different datatypes"""
        if self.value is None or other.value is None:
            return False
        else:
            try:
                return comp_func(self.value, other.value)
            except TypeError:
                return False

    def __lt__(self, other):
        return self._special_comp(other, operator.lt)

    def __eq__(self, other):
        return self._special_comp(other, operator.eq)


def safe_split(text, separator):
    """Like split for non-empty separator, list with text otherwise."""
    if separator:
        return text.split(separator)
    else:
        return [text]
